#!/usr/bin/perl -w

# WebSphere Plugin Config Merge Script
# Author: Brian D. Goodman, bdgoodman@spirirtedthought.com
# Version: 1.1
#
# Copyright Brian D. Goodman, 2008. All rights reserved.
#
# Often in working with WebSphere Application Server, there is a desire to
# create multiple profiles that can be served by a single IBM HTTP Server.
# This script assists with this task as it is currently a manual process of
# combining sections of two different plugin-cfg.xml files.
#
# This script takes three arguments: Primary config file, Secondary config file,
# and the file name of the new config file. It can be used several times to
# merge multiple plugin-cgf.xml files.
#


use strict;
use XML::LibXML;
use XML::Tidy;

# Process and cleanup out expected arguments. Do basic check to ensure we get
# the arguments we expect.

my $cfgA = $ARGV[0];
if($cfgA){chomp($cfgA)};

my $cfgB = $ARGV[1];
if($cfgB){chomp($cfgB)};

my $newcfg = $ARGV[2];
if($newcfg){chomp($newcfg)};

if(!($cfgA && $cfgB && $newcfg)){
    
    print STDOUT "\nWebSphere Plugin Config Merge Script\n";
    print STDOUT "Author: Brian D. Goodman, bdgoodman\@spirirtedthought.com\n";
    print STDOUT "Version: 1.1\n";
    print STDOUT "\n";
    
    print STDOUT ("This script takes three arguments: <primary config file>,\n");
    print STDOUT ("<secondary config file>, and <path & file name to new\n");
    print STDOUT ("config>. It can be used several times to merge multiple\n");
    print STDOUT ("plugin-cgf.xml files.\n");
    print STDOUT "\n";

    print STDOUT ("wpcMerge.pl plugin-cfg.xml plugin-cfg-other.xml");
    print STDOUT ("new-plugin-cfg.xml\n");
    
    exit(1);
}

# Merging the virtual host group section is the only fragment that needs to be
# a real merge - i.e. there is only one section per named VirtualHostGroup.

# Parse the two config files up front, they are used several times.
# Call _getVirtualHostsGroup which returns a hash of VirtualHostGroup names and
# a string representing all the VirtualHosts. This allows this script to work
# even if the config has multiple named VirtualHostGroups.

my $xmlA = XML::LibXML->new->parse_file($cfgA);
my $xmlB = XML::LibXML->new->parse_file($cfgB);

my $vhgA = &_getVirtualHostsGroup($xmlA);
my $vhgB = &_getVirtualHostsGroup($xmlB);

my %vhgA = %$vhgA;
my %vhgB = %$vhgB;

# Merge the two hash tables so we only have one hash with a unique set of
# named groups and within them unique set of virtual host entries.

my $combVHG = &_combignVHGroups(\%vhgA, \%vhgB);
my %combVHG  = %$combVHG;

# Get an XPath context so we can easily select the sections we care about,
# remove the existing VirtualHostGroups and replace them with our merged
# version.

my $xpc = XML::LibXML::XPathContext->new($xmlA);

foreach my $key ( sort keys (%$combVHG) ) {
    
    my $stmt = '/Config/VirtualHostGroup[@Name=\'' . $key . '\']';

    my @nodes = $xpc->findnodes($stmt);
    my @hosts = ();
    my $len = scalar @nodes;
    
    # We need to check len, because technically it is possible for B to have
    # VirtualHostGroups that A does not.
    if($len == 0) {
        
        my @nodes = $xpc->findnodes("/Config/VirtualHostGroup");
        my $nodeLen = scalar @nodes;

        my @vhGroups = &_returnNode($xmlB, $stmt);
        my $vhgLen = scalar @vhGroups;
        
        for(my $i=0; $i<$vhgLen; $i++){
            $nodes[$nodeLen-1]->parentNode->insertAfter($vhGroups[$i],
                                                        $nodes[$nodeLen-1]);
        }

    }
    
    
    
    for(my $i=0; $i<$len; $i++){    
        $nodes[$i]->removeChildNodes();
          
        my @values = split("!!!", $combVHG{$key});
        my $valCount = scalar @values;
        
        for(my $x=0; $x<$valCount; $x++){
          my $cnode = $nodes[$i]->addNewChild("", "VirtualHost");
          $cnode->setAttribute("Name", $values[$x]);
        }
    } 
}

# Now for the other three sections, UriGroup, Route and ServerCluster all we
# have to do is append the nodes from the second document into the first.
# To make it nice an neat I insert the nodes from the second document right
# after the nodes in the first document.

my @uriGroups = &_returnNode($xmlB, "/Config/UriGroup");
my $ugLen = scalar @uriGroups;

my @routes = &_returnNode($xmlB, "/Config/Route");
my $rLen = scalar @routes;

my @serverClusters = &_returnNode($xmlB, "/Config/ServerCluster");
my $scLen = scalar @serverClusters;

my @nodes = $xpc->findnodes("/Config/ServerCluster");
my $nodeLen = scalar @nodes;

for(my $i=0; $i<$scLen; $i++){
  $nodes[$nodeLen-1]->parentNode->insertAfter($serverClusters[$i],
                                              $nodes[$nodeLen-1]);
}

@nodes = $xpc->findnodes("/Config/UriGroup");
$nodeLen = scalar @nodes;

for(my $i=0; $i<$ugLen; $i++){
  $nodes[$nodeLen-1]->parentNode->insertAfter($uriGroups[$i],
                                              $nodes[$nodeLen-1]);
}

@nodes = $xpc->findnodes("/Config/Route");
$nodeLen = scalar @nodes;

for(my $i=0; $i<$rLen; $i++){
  $nodes[$nodeLen-1]->parentNode->insertAfter($routes[$i], $nodes[$nodeLen-1]);
}

# Lets write out the modified document to the file name and path provided as
# an argument.

# Modified to add support for Tidy
# $xmlA->toFile($newcfg, 2);

my $tidy_obj = XML::Tidy->new('xml' => $xmlA->toString(2));
$tidy_obj->tidy();
$tidy_obj->write($newcfg);



###############################################################################

# Simple helper function to return the node specified by an XPath.

sub _returnNode(){
  
  my $xml = $_[0];
  my $xpath = $_[1];
  
  my $xpc = XML::LibXML::XPathContext->new($xml);
  
  my @nodes = $xpc->findnodes($xpath);
  
  return @nodes;
}

# Helper function to combing two VirtualHostGroup hash tables.

sub _combignVHGroups(){
  
  my $a = $_[0];
  my $b = $_[1];

  my %comb = ();
  my @keysA = keys(%$a);
  my @keysB = keys(%$b);
  
  my @allkeys = (@keysA, @keysB);
  
  my $flatAllKeys = join("!!!", @allkeys);
  
  @allkeys = split("!!!", &_removeDupValues($flatAllKeys));
  
  foreach my $key (@allkeys) {
            
    if(defined $vhgA{$key} && defined $vhgB{$key}){
      $comb{$key} = &_removeDupValues($vhgA{$key} . "!!!" . $vhgB{$key});
    } elsif (defined $vhgA{$key}) {
      $comb{$key} = $vhgA{$key};
    } elsif (defined $vhgB{$key}) {
      $comb{$key} = $vhgB{$key};
    }
    
  }
  
  return \%comb;
  
}

# Pull the VirtualHostGroup(s) and then each of the vhosts, stick them in an
# array, flatten and return as part of a hash.

sub _getVirtualHostsGroup(){
  
  my $xpc = XML::LibXML::XPathContext->new($_[0]);

  my %vhgroups = ();
  
  my @nodes = $xpc->findnodes('/Config/VirtualHostGroup/@Name');
  
  foreach my $node (@nodes) {
    
    my $stmt = '/Config/VirtualHostGroup[@Name=\'' .
                $node->nodeValue . '\']/VirtualHost/@Name';
    #print "$stmt\n";
    my @hostNodes = $xpc->findnodes($stmt);
    my @hosts = ();
    my $len = scalar @hostNodes;
    
    for(my $i=0; $i<$len; $i++){
      push(@hosts, $hostNodes[$i]->nodeValue);
    }
    
    $vhgroups{$node->nodeValue} = join("!!!", @hosts);
    
  }
  
  return \%vhgroups;
}

# Helper function to remove duplicate values from a special string format.
# Values delimited by "!!!".

sub _removeDupValues(){
  my $str = $_[0];
  
  my @values = split(/!!!/, $str);
  
  my $len = @values;
  my %hash = ();
    
  for(my $i=0; $i<$len; $i++){
    $hash{$values[$i]}="";
  }
  
  return join("!!!", sort keys %hash);
}

 
